//***************************************************************************
//
//	FTDI
//
//  Example code for connecting to VNC1L Eval board and controlling a 
//  USB-missile launcher on the PIC16F688.
//
//  Version V1.00 for use with VDAP Firmware V3.63 and above
//	
//	Original Author: A.J. Dougan DADL Sept 2007
//	Refactored: G McNab FTDI Feb 2008
//	
//***************************************************************************
//	
//	Pin definitions for PIC16F688 used in this project
//	
//	PIC16F688 Pin Number		Designation			Function
//	1				VDD				Power
//	2				RA5				OSC_IN
//	3				RA4				OSC_OUT
//	4	 	        MCLR#			MCLR#
//	5				RC5				Rx (connects to TXD on VNC1L)
//	6				RC4				Tx (connects to RXD on VNC1L)
//	7				RC3				CTS# (connects to RTS# on VNC1L)
//	8				RC2				RTS# (connects to CTS# on VNC1L)
//	9				RC1				
//	10				RC0				
//	11				RA2				
//	12				RA1				
//	13			    RA0				
//	14				GND				Ground
//	
//***************************************************************************

#include <system.h>
//#include <icd2.h> // add in if using MPLAB ICD2 - allocates RAM used by ICD2

#pragma DATA 0x2007, _HS_OSC & _WDT_OFF & _PWRTE_OFF & _MCLRE_OFF & _CP_OFF & _IESO_OFF & _FCMEN_OFF;
#pragma	CLOCK_FREQ 20000000;

//***************************************************************************
//  
//  Defintions for fixed RAM location variables used in interrupt logic
//  
//***************************************************************************

char Saved_Status@0x00F0;
char Saved_W@0x00F1;

//***************************************************************************
//  
//  Defintions for Firmware Versions
//  
//***************************************************************************

#define FIRMWARE_VDAP 1
#define FIRMWARE_VDPS 2
#define FIRMWARE_VMSC1 3
#define FIRMWARE_VDFC 4

#define FIRMWARE FIRMWARE_VDAP

//***************************************************************************
//  
//  Defintions used in handling serial port
//  
//***************************************************************************

// Define serial port flags
#define _RTS  0x0002				// RTS# is RC2 (Port C bit 2)
#define _CTS  0x0003				// CTS# is RC3 (Port C bit 3)
#define FLOW_NONE  0x00	
#define FLOW_RTS_CTS  0x01
#define FIFO_EMPTY 0

// Define commonly used baud rates
#define BAUDRATE_9600 0x81
#define BAUDRATE_19200 0x40
#define BAUDRATE_57600 0x15
#define BAUDRATE_115200 0x0A

// Define serial port Rx buffer constants
#define RX_FIFO_SIZE 0x40			// Buffer size (multiple of 4 bytes)
#define RX_FIFO_OVERRUN 0x01
#define RX_FIFO_UPPER_LIMIT 0x30	// Buffer limit for handshaking
#define RX_FIFO_LOWER_LIMIT 0x20	// Buffer limit for handshaking

// Define serial port Rx buffer variables
char SERIAL_RX_FIFO[RX_FIFO_SIZE];	// Allocate buffer for RX FIFO
volatile char RX_FIFO_COUNT = 0x00;		// Reset RX FIFO count
char * RX_FIFO_HEAD_PTR = &SERIAL_RX_FIFO[0x00];	// Initialise head pointer 
char * RX_FIFO_TAIL_PTR = &SERIAL_RX_FIFO[0x00];	// Initialise tail pointer
const char * RX_FIFO_START = RX_FIFO_TAIL_PTR;			// Set lowest address of RX_FIFO
const char * RX_FIFO_END = RX_FIFO_START + RX_FIFO_SIZE;	// Set highest address of RX_FIFO
char RX_STATUS = 0x00;				// Reset RX status
char FLOW_CONTROL_ENABLED = 0x00;	// Flow control settings

//***************************************************************************
//  
//  Defintions of constants and variables for timer
//  
//***************************************************************************

// Define timer
char timerCount = 0;				// millisecond count for timer
char timerStatus = 0;				// status of timer
#define ReqTimer 0x01				// timerStatus: Bit 0 - timer enabled
#define DoneTimer 0x02				// timerStatus: Bit 1 - timer expired

//***************************************************************************
//  
//  Defintions of constants and variables for main code
//  
//***************************************************************************

// Define USB device activity to detect Device Detected or Removed messages (bitmap)
unsigned char usbStatus = 0;
#define usbDeviceDet1 0x01			// usbStatus: Bit 0 - launcher 
                                    //  detected/removed on USB Port 1
#define usbDeviceDet2 0x02			// usbStatus: Bit 1 - launcher 
                                    //  detected/removed on USB Port 2

// Button and LED state for main loop
enum tpAction {
    Action_Idle,
    Action_Left,
    Action_Right,
    Action_Up,
    Action_Down,
    Action_Fire,
	Action_NoDev,
};

// Prompts returned by VNC1L
enum tpPrompt {
    Prompt_None,
    Prompt_OK,	// > or ND (Success)
    Prompt_UE,	// E echo
    Prompt_LE,	// e echo
    Prompt_CF,	// CF (Command Failed)
    Prompt_BC,	// BC (Bad Command)
    Prompt_DF,	// DF (Disk Full)
    Prompt_FI,	// FI (File Invalid)
    Prompt_RO,	// RO (Read Only)
    Prompt_FO,	// FO (File Open)
    Prompt_NE,	// NE (Dir Not Empty)
    Prompt_FN,	// FN (Filename Invalid)
    Prompt_END, // End of list marker
};

// String definitions for prompts 
// ROM arrays can be a maximum of 255 characters long and must be addressed
// using [] operators.
// The format of this array is prompt code, number of characters in prompt,
// then the text of the prompt to match.
// ROM reads are comparitively slow so this array is arranged to reduce
// the number of ROM reads required to determine the prompt code.
#define MAX_PROMPT_LEN 8
rom char *pPrompt	= {
	// SCS Prompts
    Prompt_OK, 2, '>',0x0d,
    Prompt_OK, 3, 'N','D',0x0d,
    Prompt_UE, 2, 'E',0x0d,
    Prompt_LE, 2, 'e',0x0d,
    Prompt_CF, 3, 'C','F',0x0d,
    Prompt_BC, 3, 'B','C',0x0d,
    Prompt_DF, 3, 'D','F',0x0d,
    Prompt_FI, 3, 'F','I',0x0d,
    Prompt_RO, 3, 'R','O',0x0d,
    Prompt_FO, 3, 'F','O',0x0d,
    Prompt_NE, 3, 'N','E',0x0d,
    Prompt_FN, 3, 'F','N',0x0d,
	// ECS Prompts
    Prompt_OK, 5, 'D',':','\\','>',0x0d,
    Prompt_OK, 8, 'N','o',' ','D','i','s','k',0x0d,
    Prompt_END, 0,
};

// Asynchronous messages returned by VNC1L
enum tpMessage {
    Message_None,
    Message_NU,		// NU / No Upgrade
    Message_DD1,	// DD1 / Device Detected USB Port 1)
    Message_DD2,	// DD2 / Device Detected USB Port 2)
    Message_DR1,	// DR1 / Device Removed USB Port 1)
    Message_DR2,	// DR2 / Device Removed USB Port 2)
#if FIRMWARE == FIRMWARE_VDPS
	Message_SDA,	// SDA / Slave Device Active on VDPS)
	Message_SDD,	// SDD  Slave Device Disabled on VDPS)
#endif
    Message_END,	// End of list marker
};

// String definitions for messages 
// The format of this array is message code, number of characters in message,
// then the text of the message to match.
#define MAX_MESSAGE_LEN 20
rom char *pMessage	= {
	// SCS Messages
	Message_DD1, 4, 'D','D','1',0x0d,
    Message_DD2, 4, 'D','D','2',0x0d,
    Message_DR1, 4, 'D','R','1',0x0d,
    Message_DR2, 4, 'D','R','2',0x0d,
    Message_NU, 3, 'N','U',0x0d,
#if FIRMWARE == FIRMWARE_VDPS
	Message_SDA, 4, 'S','D','A',0x0d,
	Message_SDD, 4, 'S','D','D',0x0d,
#endif
    // ECS messages are required to detect devices before the SCS command
    // has been sent - the Vinculum Firmware Customiser tool can be used
    // to start the VNC1L in SCS mode by default
    Message_DD1, 19, 'D','e','v','i','c','e',' ',
                    'D','e','t','e','c','t','e','d',' ','P','1',0x0d,
    Message_DD2, 19, 'D','e','v','i','c','e',' ',
                    'D','e','t','e','c','t','e','d',' ','P','2',0x0d,
    Message_DR1, 18, 'D','e','v','i','c','e',' ',
                    'R','e','m','o','v','e','d',' ','P','1',0x0d,
    Message_DR2, 18, 'D','e','v','i','c','e',' ',
                    'R','e','m','o','v','e','d',' ','P','2',0x0d,
    Message_NU, 11, 'N','o',' ','U','p','g','r','a','d','e',0x0d,
#if FIRMWARE == FIRMWARE_VDPS
	Message_SDA, 14, 'S','l','a','v','e',' ',
                    'E','n','a','b','l','e','d',0x0d,
	Message_SDD, 15, 'S','l','a','v','e' ',
                    'D','i','s','a','b','l','e','d',0x0d,
#endif
    Message_END, 0,
};

//***************************************************************************
//  
//  Defintions of constants used by VNC1L
//  
//***************************************************************************

// VNC1L addresses for accessing I/O Ports in IOR/IOW commands
#define PortAD 0x00
#define PortAC 0x01
#define PortBD 0x02
#define PortBC 0x03

// VNC1L monitor command defintions for Short Command Set
#define CMD_IOW 0x2A					// IOW command
#define CMD_IOR 0x29					// IOR command
#define CMD_QueryDevice 0x85			// QD command
#define CMD_QueryPort1 0x2B				// QP1 command
#define CMD_QueryPort2 0x2C				// QP2 command
#define CMD_SetCurrentDev 0x86			// SC command
#define CMD_SendSetup 0x9A				// SSU command

// VNC1L direction masks for I/O Ports in IOW command
#define PortAC_Dir 0x20					// Outputs - Bit 5
#define PortBD_Dir 0xF0					// Outputs - Bits 7-4

// VNC1L bit definitions for buttons on BD I/O Port
#define PortBD_Button_Left 0x01			// Bit 0 - 0x01
#define PortBD_Button_Right 0x02		// Bit 1 - 0x02
#define PortBD_Button_Up 0x04			// Bit 2 - 0x04
#define PortBD_Button_Down 0x08			// Bit 3 - 0x08

// VNC1L bit definitions for button on AC I/O Port
#define PortAC_Button_Fire 0x40			// Bit 6 - 0x40

// VNC1L bit definitions for LEDs on BD I/O Port
#define PortBD_Led_Left 0x10			// Bit 4 - 0x10
#define PortBD_Led_Right 0x20			// Bit 5 - 0x20
#define PortBD_Led_Up 0x40				// Bit 6 - 0x40
#define PortBD_Led_Down 0x80			// Bit 7 - 0x80
#define PortBD_Led_All 0xF0				// Bits 4- 7 - 0xF0

// VNC1L bit definitions for LED on AC I/O Port
#define PortAC_Led_Fire 0x20			// Bit 5 - 0x20


//***************************************************************************
//  
//  Define fixed data for launcher control in ROM space
//  
//***************************************************************************

// SSU command string defintions for USB commands sent to the rocket launcher 
// to control it's operation.
#define SSU_CMD_LEN 17
#define SSU_PAD_LEN 56

// Offsets in the SSU command string for each command 
#define SSU_DATA1 0
#define SSU_DATA2 1
#define SSU_DATA_LEFT 2
#define SSU_DATA_RIGHT 3
#define SSU_DATA_UP 4
#define SSU_DATA_DOWN 5
#define SSU_DATA_FIRE 6
#define SSU_DATA_STOP 7

// SSU command string
// ROM arrays can be a maximum of 255 characters long and must be addressed
// using [] operators.
rom char *ssuData = {
	// SSU_DATA1
	0x21,0x09,0x00,0x02,0x01,0x00,0x08,0x00,0x0D,
	0x55,0x53,0x42,0x43,0x00,0x00,0x04,0x00,
	// SSU_DATA2
	0x21,0x09,0x00,0x02,0x01,0x00,0x08,0x00,0x0D,
	0x55,0x53,0x42,0x43,0x00,0x40,0x02,0x00,
	// SSU_DATA_LEFT
	0x21,0x09,0x00,0x02,0x00,0x00,0x40,0x00,0x0D,
	0x00,0x01,0x00,0x00,0x00,0x00,0x08,0x08,
	// SSU_DATA_RIGHT
	0x21,0x09,0x00,0x02,0x00,0x00,0x40,0x00,0x0D,
	0x00,0x00,0x01,0x00,0x00,0x00,0x08,0x08,
	// SSU_DATA_UP
	0x21,0x09,0x00,0x02,0x00,0x00,0x40,0x00,0x0D,
	0x00,0x00,0x00,0x01,0x00,0x00,0x08,0x08,
	// SSU_DATA_DOWN
	0x21,0x09,0x00,0x02,0x00,0x00,0x40,0x00,0x0D,
	0x00,0x00,0x00,0x00,0x01,0x00,0x08,0x08,
	// SSU_DATA_FIRE
	0x21,0x09,0x00,0x02,0x00,0x00,0x40,0x00,0x0D,
	0x00,0x00,0x00,0x00,0x00,0x01,0x08,0x08,
	// SSU_DATA_STOP
	0x21,0x09,0x00,0x02,0x00,0x00,0x40,0x00,0x0D,
	0x00,0x00,0x00,0x00,0x00,0x00,0x08,0x08,
};

//***************************************************************************
//  
//  Define function prototypes
//  
//***************************************************************************

// serial port interface low-level routines
void interrupt();
void serialInit(unsigned char baudRate, char flowCtl);
void serialResetRxFifo();
void serialAddToRxFifo();
unsigned char serialRxFifoStatus();
unsigned char serialReadFromRxFifo();
void serialSendByte(char txByte);

// serial port interface routines
enum tpMessage serialWaitForData();
enum tpPrompt serialWaitForPrompt();
char serialWaitForCR();

// timer routines
void initTimer0();
void msDelay(char count);

// monitor operations
void monSync();
void monSetLEDs(enum tpAction uAction);
enum tpAction monReadButtons();
void monWaitForButtonRelease();

// USB routines
void monShrtCmdSet();
int usbSelectLauncher();
void usbSendSSU(int offset);
void usbSendStartSeq();

//****************************************************************************
//
// Interrupt routine
//
//****************************************************************************

void interrupt()
{
    //-------------------------------
    // save status and W register
    //-------------------------------
    asm
    {
        MOVWF _Saved_W;	Copy W to TEMP register
        SWAPF _status,W;	Swap status to be saved into W
        ;Swaps are used because they do not affect the status bits
        MOVWF _Saved_Status;
    }
    //-------------------------------
    //
    //-------------------------------

    if (pir1.RCIF) 
    {
        // Add received byte to Rx FIFO
        serialAddToRxFifo();
        // Clear RCIF flag
        pir1.RCIF = 0;
    }
    if (intcon.T0IF)
    {
        // check if we are timing anything
        if (timerStatus.ReqTimer)
        {
            timerCount --;
            if (!timerCount )
            {
                timerStatus.DoneTimer = 1;
                timerStatus.ReqTimer = 0;
                // reload timer to give approx 1ms
                tmr0 = 0xED;
            }
        }
        // clear TMR0 IRQ Flag
        intcon.T0IF = 0;
    }	

    //-------------------------------
    // restore status and W register
    //-------------------------------
    asm
    {
        SWAPF _Saved_Status,W;	Swap STATUS_TEMP register into W
        ;   (sets bank to original state)
        MOVWF _status;			Move W into STATUS register
        SWAPF _Saved_W,F ;		Swap W_TEMP
        SWAPF _Saved_W,W ;		Swap W_TEMP into W
    }
}

//***************************************************************************
// Name: serialResetRxFifo
//
// Description: Initialise the serial UART interface. 
//
// Parameters: None.
//
// Returns: None.
//
// Comments: Set baud rate, handshaking, interrupt enable and finally
//  enable the serial UART receiver.
//
//***************************************************************************

void serialInit(unsigned char baudRate, char flowCtl)
{
    // Disable interrupts
    pie1.RCIE = 0;											
    intcon.GIE = 0;
    intcon.PEIE = 0;

    baudctl = 0x00;	
    // Digital IO on RC2, RC3
    ansel &= 0x3F;											
    // Set RC4 (Tx) & RC2 (RTS#) as output,RC5 (Rx) & RC3 (CTS#) as input
    trisc = 0xDB;											
    // TRISC bits 1 for input, 0 for output
    // Set RTS# high (inactive) at this stage
    portc._RTS = 1;											
    // Enable UART module and set 8,n,1 - Set BRGH to 1
    txsta = 0x04;											
    // Enable serial port
    rcsta.SPEN = 1;											

    spbrg = baudRate;
    spbrgh = 0x00;			

    // Set flow control enabled flag
    FLOW_CONTROL_ENABLED = flowCtl;				

    if (flowCtl == FLOW_RTS_CTS)
    {
        // If handshaking is required, set RTS# active (low)
        portc._RTS = 0;										
    }

    // Reset Rx FIFO
    serialResetRxFifo();									

    // Clear Rx interrupt flag
    pir1.RCIF = 0;											
    // Enable interrupts for Rx byte
    pie1.RCIE = 1;											
    intcon.GIE = 1;
    intcon.PEIE = 1;
    // Enable receiver
    rcsta.CREN = 1;											
}

//***************************************************************************
// Name: serialResetRxFifo
//
// Description: Reset the serial UART interface and the FIFO buffer. 
//
// Parameters: None.
//
// Returns: None.
//
// Comments: 
//
//***************************************************************************

void serialResetRxFifo()
{
    // Reset FIFO count to 0
    RX_FIFO_COUNT = 0x00;
    // Reset FIFO status
    RX_STATUS = 0x00;
    // Set FIFO head pointer to address of first byte of SERIAL_RX_FIFO
    RX_FIFO_HEAD_PTR = (&SERIAL_RX_FIFO[0x00]);
    // Set FIFO tail pointer to address of first byte of SERIAL_RX_FIFO
    RX_FIFO_TAIL_PTR = (&SERIAL_RX_FIFO[0x00]);
}

//***************************************************************************
// Name: serialAddToRxFifo
//
// Description: Retrieve data from the serial UART interface and add it to
//  the FIFO buffer. 
//
// Parameters: None.
//
// Returns: None.
//
// Comments: This is called by the interrupt routine only. This routine will
//  set RTS if the size of the FIFO buffer exceeds the upper limit threshold.
//
//***************************************************************************

void serialAddToRxFifo()
{
    if (RX_FIFO_COUNT >= RX_FIFO_SIZE)						// If Rx FIFO is full
    {
        RX_STATUS = RX_STATUS | RX_FIFO_OVERRUN;			// Set overrun flag
        if (FLOW_CONTROL_ENABLED == FLOW_RTS_CTS)			// If using flow control
        {
            portc._RTS = 1;									// Set RTS# inactive (high)
        }
        return;
    }

    if (RX_FIFO_COUNT >= RX_FIFO_UPPER_LIMIT)				// Keep some space in the FIFO for bytes already being sent
    {
        if (FLOW_CONTROL_ENABLED == FLOW_RTS_CTS)			// If using flow control
        {
            portc._RTS = 1;									// Set RTS# inactive (high)
        }
    }


    *RX_FIFO_HEAD_PTR++ = rcreg;							// Add received byte to RX_FIFO and then increment head pointer
    RX_FIFO_COUNT++;										// Increment byte count

    if (RX_FIFO_HEAD_PTR >= RX_FIFO_END)					// If head pointer is greater than allocated FIFO end address
    {
        RX_FIFO_HEAD_PTR = RX_FIFO_START;					// Wrap-around buffer to start address again (circuluar FIFO)
    }	

    if ((RX_FIFO_COUNT < RX_FIFO_UPPER_LIMIT) && (FLOW_CONTROL_ENABLED == FLOW_RTS_CTS))	// If Rx FIFO is not full
    {
        portc._RTS = 0;										// Set RTS# active (low)
    }

}

//***************************************************************************
// Name: serialRxFifoStatus
//
// Description: Find the size of the serial UART FIFO buffer to determine if
//  there is data waiting in the buffer to be processed.
//
// Parameters: None.
//
// Returns: Number of characters in the FIFO buffer (zero if empty).
//
// Comments: This should be non-zero before reading any data from the serial
//  UART FIFO.
//
//***************************************************************************

unsigned char serialRxFifoStatus()
{
    return RX_FIFO_COUNT;
}

//***************************************************************************
// Name: serialReadFromRxFifo
//
// Description: Return a single byte of data from the serial UART FIFO
//  buffer.
//
// Parameters: None.
//
// Returns: First character from the FIFO buffer.
//
// Comments: The RTS line may be activated by this routine when the data
//  taken from the FIFO buffer brings it's size below the lower limit
//  threshold.
//
//***************************************************************************

unsigned char serialReadFromRxFifo()
{
    unsigned char rxFifoByte;

    // Disable interrupts
    intcon.GIE  = 0;								

    // Read a byte from the FIFO at position given by RX_FIFO_TAIL_PTR and then increment tail pointer
    rxFifoByte = *RX_FIFO_TAIL_PTR++;				
    // Decrement byte count	
    RX_FIFO_COUNT--;								

    // If tail pointer is greater than allocated FIFO end address
    if (RX_FIFO_TAIL_PTR >= RX_FIFO_END)			
    {
        // Wrap-around buffer to start address again (circuluar FIFO)
        RX_FIFO_TAIL_PTR = RX_FIFO_START;			
    }

    // If using flow control
    if ((RX_FIFO_COUNT < RX_FIFO_LOWER_LIMIT) && (FLOW_CONTROL_ENABLED == FLOW_RTS_CTS))	
    {
        // Set RTS# active (low)
        portc._RTS = 0;								
    }

    // Re-enable interrupts
    intcon.GIE  = 1;								

    return rxFifoByte;
}

//***************************************************************************
// Name: serialSendByte
//
// Description: Send a single byte of data to the serial UART.
//
// Parameters: Character to send.
//
// Returns: None.
//
// Comments: The data is put into the transmit buffer and sent immediately 
//  once the the CTS line is active.
//
//***************************************************************************

void serialSendByte(char txByte)
{
    if (FLOW_CONTROL_ENABLED == FLOW_RTS_CTS)
    {	
        // Wait for CTS# to become active (low)
        while (portc._CTS== 1);						
    }
    // Wait for TSR to be empty
    while (!txsta.TRMT);							
    // Move byte to send into TXREG
    txreg = txByte;									
    // Start transmission
    txsta.TXEN = 1;									
    // Wait for TSR to be empty - indicates sending is complete
    while (!txsta.TRMT);							
}

//***************************************************************************
// Name: serialWaitForData
//
// Description: Waits until a line of data has been received and available in 
//  the receive buffer. Also peeks at the receive buffer to match an
//  asynchronous information message. This is always terminated by a carriage
//  return. Data is matched against a ROM image of possible messages.
//
// Parameters: None.
//
// Returns: enum tpMessage with the last detected message.
//
// Comments: This will return when data from a command has been received and
//  any messages have been filtered out. The response from a command must be
//  be terminated by a carriage return.
//
//***************************************************************************

enum tpMessage serialWaitForData()
{
    char cThis = 0;
    char cBuf[MAX_MESSAGE_LEN];
    int romBufOffset;
    int bufCnt, romCnt;
    enum tpMessage tpResp, tpThis;
    int i, j;
    unsigned char rxFifoByte;
	unsigned char *rxFifoPtr;

	while (1)
	{
		// detect a line (terminated with a carriage return) in
		// the receive FIFO
		bufCnt = 0;

		// assuming that there are characters to be received
		if (RX_FIFO_COUNT)
		{
			intcon.GIE  = 0;								
			rxFifoPtr = RX_FIFO_TAIL_PTR;
			intcon.GIE  = 1;

			while (1)
			{
				rxFifoByte = *rxFifoPtr++;
				// add to message buffer
				cBuf[bufCnt++] = rxFifoByte;

				// carriage return received
				if (rxFifoByte == 0x0d)	
				{
					break;
				}

				// protect message buffer
				if (bufCnt == MAX_MESSAGE_LEN)
				{
					// the received data in the buffer is too long to be a 
					// message, as such there is no point in continuing
					return Message_None;
				}

				intcon.GIE  = 0;
				// ff tail pointer is greater than allocated FIFO end address
				if (rxFifoPtr >= RX_FIFO_END)
				{
					// wrap-around buffer to start address again (circuluar FIFO)
					rxFifoPtr = RX_FIFO_START;
				}
				intcon.GIE  = 1;

				// delay until a further byte is received
				while (RX_FIFO_COUNT == bufCnt);
			}
		}

		// if there is a carriage return in the buffer
		if (bufCnt)
		{
			// initialise prompt detection
			tpResp = Message_None;
			romBufOffset = 0;

			// find if the Message matches any of our predefined types
			while (tpResp == Message_None)
			{
				// message code for this entry
				tpThis = pMessage[romBufOffset++];

				// number of characters for this entry
				romCnt = pMessage[romBufOffset++];

				// reached end of Message definitions
				if (romCnt == 0)
				{
					tpResp = Message_None;
					break;
				}

				// check against this Message if length is correct
				if (romCnt == bufCnt)
				{
					tpResp = tpThis;

					for (j = 0; j < bufCnt; j++)
					{
						cThis = pMessage[romBufOffset++];
						romCnt--;

						// mis-match of data
						if (cThis != cBuf[j])
						{
							tpResp = Message_None;
							break; // break from for loop
						}
					}
				}

				romBufOffset += romCnt;
			} // while (tpResp == Message_None)

			// if a message was received then act
			if (tpResp != Message_None)
			{
				// delete acknowledged message from FIFO
				for (i = 0; i < bufCnt; i++)
				{
					serialReadFromRxFifo();
				}

				// perform action depending on the message
				// ignore message which have no actions
				if (tpResp == Message_DD1)
				{
					usbStatus.usbDeviceDet1 = 1;
				}
				if (tpResp == Message_DD2)
				{
					usbStatus.usbDeviceDet2 = 1;
				}
				if (tpResp == Message_DR1)
				{
					usbStatus.usbDeviceDet1 = 0;
				}
				if (tpResp == Message_DR2)
				{
					usbStatus.usbDeviceDet2 = 0;
				}
			}
			else
			{
				// non-message response received
				break;
			}
		}
	// loop to check for more messages
	};

	return tpResp;
}

//***************************************************************************
// Name: serialWaitForPrompt
//
// Description: Read data from the serial UART until a carriage return is
//  received. Decode the data received and determine the prompt from the 
//  Vinculum monitor.
//
// Parameters: None.
//
// Returns: enum tpPrompt with the detected prompt.
//
// Comments: There is a timeout implemented using the timer. If a "Device
//  Detected" or "No Update" message is received then the routine will 
//  retry and wait for a completion prompt.
//
//***************************************************************************

enum tpPrompt serialWaitForPrompt()
{
    char cThis = 0;
    char cBuf[MAX_PROMPT_LEN];
    int romBufOffset;
    int bufCnt, romCnt;
    enum tpPrompt tpResp, tpThis;
    int i, j;

    while (1)
    {
        // initialise prompt compare buffer 
        bufCnt = 0;

        // wait for prompt to be received (break if timer elapses)
        do
        {
            // wait for data to be received
            while (RX_FIFO_COUNT == FIFO_EMPTY);

			// character available
            cThis = serialReadFromRxFifo();
            
            // ignore any other characters sent after prompt
            if (bufCnt < MAX_PROMPT_LEN)
                cBuf[bufCnt++] = cThis;

        } while (cThis != 0x0d);

        // initialise prompt detection
        tpResp = Prompt_None;
        romBufOffset = 0;

        // only check if there are characters in the buffer to check
        if (bufCnt)
        {
            // find if the prompt matches any of our predefined types
            while (tpResp == Prompt_None)
            {
                // prompt code for this entry
                tpThis = pPrompt[romBufOffset++];

                // number of characters for this entry
                romCnt = pPrompt[romBufOffset++];

                // reached end of prompt definitions
                if (romCnt == 0)
                {
                    tpResp = Prompt_None;
                    break;
                }

                // check against this prompt if length is correct
                if (romCnt == bufCnt)
                {
                    tpResp = tpThis;

                    for (j = 0; j < bufCnt; j++)
                    {
                        cThis = pPrompt[romBufOffset++];
                        romCnt--;

                        // mis-match of data
                        if (cThis != cBuf[j])
                        {
                            tpResp = Prompt_None;
                            break; // break from for loop
                        }
                    }
                }

                romBufOffset += romCnt;
            }
        }

        // valid prompt received
        break;
    }

    return tpResp;
}

//***************************************************************************
// Name: serialWaitForCR
//
// Description: Read data from the serial UART until a carriage return is
//  received.
//
// Parameters: None.
//
// Returns: Character immediately preceeding the carriage return.
//
// Comments: There is a timeout implemented using the timer.
//
//***************************************************************************

char serialWaitForCR()
{
    char cThis = 0, cLast;

    // wait for prompt to be received (break if timer elapses)
    do
    {
        cLast = cThis;
        // wait for data to be received
        while (RX_FIFO_COUNT == FIFO_EMPTY);

		// character available
        cThis = serialReadFromRxFifo();

    } while (cThis != 0x0d);

    return cLast;
}

//***************************************************************************
// Name: initTimer0
//
// Description: Initialise and start the timer.
//
// Parameters: None.
//
// Returns: None.
//
// Comments: Timer may be used for a timeout when detecting a response or
//  as a method for implementing a delay.
//  Timer not used at present.
//
//***************************************************************************

void initTimer0()
{
    intcon.T0IF = 0;
    // timer to give approx 1ms
    tmr0 = 0xED;					
    // with prescaler set to divide by 256
    option_reg = 0xD7;				
    intcon.T0IE = 1;

    timerStatus.DoneTimer = 0;
    timerStatus.ReqTimer = 0;
    timerCount = 0x00; 
}

//***************************************************************************
// Name: msDelay
//
// Description: Waits for a specified number of ticks of the timer.
//
// Parameters: Number of timer ticks to delay (approx 1ms units).
//
// Returns: None.
//
// Comments: None.
//
//***************************************************************************

void msDelay(char count)
{
    timerStatus.DoneTimer = 0;
    timerCount = count;
    timerStatus.ReqTimer = 1;
    while ((timerStatus.DoneTimer == 0));
}

//***************************************************************************
// Name: monSync
//
// Description: Synchronises with the Vinclulum monitor.
//
// Parameters: None.
//
// Returns: None.
//
// Comments: Sends an echo command 'E' to the Vinculum and waits for a valid
//  response. Then sends an echo command 'e' to verify.
//
//***************************************************************************

void monSync()
{
    while (1)
    {
        serialSendByte('E');
        serialSendByte(0x0D);

		serialWaitForData();
        if (serialWaitForPrompt() == Prompt_UE)
        {
            serialSendByte('e');
            serialSendByte(0x0D);

			serialWaitForData();
            if (serialWaitForPrompt() == Prompt_LE)
            {
                break;
            }
        }

        // flush RXbuffer
		msDelay(20);
        serialResetRxFifo();		
    }
}

//***************************************************************************
// Name: monSetLEDs
//
// Description: Waits for all buttons to be released before sending a stop
//  command to the launcher.
//
// Parameters: enum tpAction indicating which LEDs should be active.
//
// Returns: None.
//
// Comments: Sets or clears the 5 LEDs on the V-EVAL development board which 
//  are on Port AC and BD. These are active low. 
//
//***************************************************************************

void monSetLEDs(enum tpAction uAction)
{
    unsigned char uLedAC = 0, uLedBD = 0;

    // set or clear fire LED first
    // bit is low to turn on LED or high to turn off
    // clear selected LED bit depending on action
    switch (uAction)
    {
    case Action_Fire:
        uLedAC = PortAC_Led_Fire; break;
    case Action_Left:
        uLedBD = PortBD_Led_Left; break;
    case Action_Right:
        uLedBD = PortBD_Led_Right; break;
    case Action_Up:
        uLedBD = PortBD_Led_Up; break;
    case Action_Down:
        uLedBD = PortBD_Led_Down; break;
	case Action_NoDev:
        uLedBD = PortBD_Led_All; break;
    }

    uLedAC = ~uLedAC;
    uLedBD = ~uLedBD;

    // send IOW command to write to port AC
    serialSendByte(CMD_IOW);
    serialSendByte(' ');
    serialSendByte(PortAC);
    serialSendByte(PortAC_Dir);
    serialSendByte(uLedAC);
    serialSendByte(0x0D);

	serialWaitForData();
    serialWaitForPrompt();

    // send IOW command to write to port BD
    serialSendByte(CMD_IOW);
    serialSendByte(' ');
    serialSendByte(PortBD);
    serialSendByte(PortBD_Dir);
    serialSendByte(uLedBD);
    serialSendByte(0x0D);

	serialWaitForData();
    serialWaitForPrompt();
}

//***************************************************************************
// Name: monReadButtons
//
// Description: Waits for all buttons to be released before sending a stop
//  command to the launcher.
//
// Parameters: None.
//
// Returns: enum tpAction corresponding to buttons pressed.
//
// Comments: Reads the 5 buttons on the V-EVAL development board which are
//  on Port AC and BD. These are active low. Fire button has highest 
//  priority.
//
//***************************************************************************

enum tpAction monReadButtons()
{
    unsigned char uButtonAC, uButtonBD;
    enum tpAction uAction;

    // send IOR command to read from port AC
    serialSendByte(CMD_IOR);
    serialSendByte(' ');
    serialSendByte(PortAC);
    serialSendByte(0x0D);

	// wait for 1 byte of data to be received
	serialWaitForData();
    // read fire button status
    uButtonAC = serialReadFromRxFifo();
    uButtonAC = ~uButtonAC;

	// next character is a carriage return
	serialReadFromRxFifo();
    serialWaitForPrompt();

    // send IOR command to read from port AC
    serialSendByte(CMD_IOR);
    serialSendByte(' ');
    serialSendByte(PortBD);
    serialSendByte(0x0D);

	// wait for 1 byte of data to be received
	serialWaitForData();
    // now read direction button status
    uButtonBD = serialReadFromRxFifo();
    uButtonBD = ~uButtonBD;

	// next character is a carriage return
	serialReadFromRxFifo();
    serialWaitForPrompt();

    // fire button will take precedence over direction buttons
    if (uButtonAC & PortAC_Button_Fire)
    {
        uAction = Action_Fire;
    }
    // set action depending on direction button
    else if (uButtonBD & PortBD_Button_Left)
    {
        uAction = Action_Left;
    }
    else if (uButtonBD & PortBD_Button_Right)
    {
        uAction = Action_Right;
    }
    else if (uButtonBD & PortBD_Button_Up)
    {
        uAction = Action_Up;
    }
    else if (uButtonBD & PortBD_Button_Down)
    {
        uAction = Action_Down;
    }
    else
    {
        // default is no button pressed
        uAction = Action_Idle;
    }

    return uAction;
}

//***************************************************************************
// Name: monWaitForButtonRelease
//
// Description: Waits for all buttons to be released before sending a stop
//  command to the launcher.
//
// Parameters: None.
//
// Returns: None.
//
// Comments: Prompt handling for the aim and fire commands is performed in
//  this routine.
//
//***************************************************************************

void monWaitForButtonRelease()
{
    // read buttons until none are pressed
    while(monReadButtons() != Action_Idle);
}

//***************************************************************************
// Name: monShrtCmdSet
//
// Description: Switches the Vinculum monitor to Short Command Set mode.
//
// Parameters: None.
//
// Returns: None.
//
// Comments: The mode may be changed by editing the firmware using the 
//  Vinculum Firmware Customiser tool.
//
//***************************************************************************

void  monShrtCmdSet()
{
    serialSendByte('s');
    serialSendByte('c');
    serialSendByte('s');
    serialSendByte(0x0D);

	serialWaitForData();
    serialWaitForPrompt();
}

//***************************************************************************
// Name: usbSelectLauncher
//
// Description: Queries the 16 possible devices on the Vinculum and selects
//  the first launcher detected.
//
// Parameters: None.
//
// Returns: 1 for a launcher detected or zero for no launcher.
//
// Comments: Using the QD command, low-speed HID devices are found and if 
//  the VID and PID match the launcher then it is selected with the SC
//  command.
//
//***************************************************************************

int usbSelectLauncher()
{
    unsigned char uType;
    unsigned char devInfo[32];
    int dev, i;
    int status = 0;

    // Send QD 0 to detect device on USB Port
    // a rocket launcher will return:
    // xx 08 81 08 00 00 03 08 01 02 00 03 00 00 30 11 
    // 02 02 00 01 02 00 00 00 00 00 00 00 00 00 00 00 <cr>

	// NOTE: a suggested improvement would be to loop through all 16 
	// possible devices and identify and select the first launcher

	// only check "QD 0" slot - assumes launcher is the only USB device
	// currently connected
	dev = 0;
    serialSendByte(CMD_QueryDevice);
    serialSendByte(' ');
    serialSendByte(dev);
    serialSendByte(0x0D);

	serialWaitForData();
    for (i = 0; i < 32; i++)
    {
        while (serialRxFifoStatus() == FIFO_EMPTY);
        devInfo[i] = serialReadFromRxFifo();
    }
    serialWaitForCR();
    serialWaitForPrompt();

    // check we are looking at a device that it is a low speed HID device
    if ((devInfo[20] == 0x02) && (devInfo[7] == 0x08))
    {
        // check the device info matches rocket launcher's VID & PID
        if ((devInfo[14] == 0x30) && (devInfo[15] == 0x11) &&
            (devInfo[16] == 0x02) && (devInfo[17] == 0x02))
        {
			// connect to launcher on USB Port 1 with SC command
			serialSendByte(CMD_SetCurrentDev);
			serialSendByte(' ');
			serialSendByte(dev);
			serialSendByte(0x0D);

			serialWaitForData();
			serialWaitForPrompt();
	        
			// successfully found and connected to the device
			status = 1;
		}
    }

    return status;
}

//***************************************************************************
// Name: usbSendStartSeq
//
// Description: Sends two SSU commands to the launcher as a preamble to an
//  aim or fire command.
//
// Parameters: None.
//
// Returns: None.
//
// Comments: Data for the first SSU_CMD_LEN bytes of the SSU command is held 
//  in ROM. No padding required.
//
//***************************************************************************

void usbSendStartSeq()
{
    int i;

    serialSendByte(CMD_SendSetup);
    serialSendByte(' ');
    for (i=0; i < SSU_CMD_LEN; i++)
    {
        serialSendByte(ssuData[(SSU_DATA1 * SSU_CMD_LEN) + i]);
    }
	serialWaitForData();
    serialWaitForPrompt();

    serialSendByte(CMD_SendSetup);
    serialSendByte(' ');
    for (i=0; i < SSU_CMD_LEN; i++)
    {
        serialSendByte(ssuData[(SSU_DATA2 * SSU_CMD_LEN) + i]);
    }
	serialWaitForData();
    serialWaitForPrompt();
}

//***************************************************************************
// Name: usbSendSSU
//
// Description: Sends an SSU commands using the data in the SSU DATA rom.
//
// Parameters: None.
//
// Returns: None.
//
// Comments: Data for the first SSU_CMD_LEN bytes of the SSU command is held 
//  in ROM. No padding required.
//
//***************************************************************************

void usbSendSSU(int offset)
{
    int i;

	usbSendStartSeq();
    serialSendByte(CMD_SendSetup);
    serialSendByte(' ');
    for (i = 0; i < SSU_CMD_LEN; i++)
    {
        serialSendByte(ssuData[(offset * SSU_CMD_LEN) + i]);
    }
    for (i = 0; i < SSU_PAD_LEN; i++) serialSendByte(0x00);
	serialWaitForData();
    serialWaitForPrompt();
}

//***************************************************************************
// Name: main
//
// Description: Main loop for program. Initialises PIC UART interface to 
//  Vinculum then calls monSync() until a launcher is detected. Once this
//  happens it will poll the control buttons, sending commands to the 
//  launcher and LEDs until the launcher is removed.
//
// Parameters: None.
//
// Returns: None.
//
// Comments: tpaStatus holds the current state of the buttons. usbStatus
//  holds the connection status of devices on the 2 USB ports on the 
//  Vinculum.
//
//***************************************************************************

void main()
{
    enum tpAction tpaStatus;
    unsigned char usbPrevStatus;
    int usbLauncher;

    // set initial status as unconnected and force a scan for a launcher
    usbStatus = 0;
    usbPrevStatus = !usbStatus;

    // initialise Baud rate and flow control
    serialInit(BAUDRATE_9600, FLOW_RTS_CTS);
    initTimer0();

    // wait for the Monitor to become active
    // first data from the Vinculum is "VcccVxx.xx On-Line:"
    while (serialWaitForCR() != ':');

    // configure and setup monitor interface
    monSync();
    monShrtCmdSet();
    monSetLEDs(Action_NoDev);
    
    while (1)
    {
        if (usbStatus != usbPrevStatus)
        {
            usbPrevStatus = usbStatus;
            // attempt to find a launcher
            usbLauncher = usbSelectLauncher();
        }

        // verify and select the launcher device
        if (usbLauncher)
        {
            tpaStatus = monReadButtons();
            monSetLEDs(tpaStatus);

            if(tpaStatus == Action_Left) 
            {
				usbSendSSU(SSU_DATA_LEFT);
            }
            else if(tpaStatus == Action_Right) 
            {
				usbSendSSU(SSU_DATA_RIGHT);
            }
            else if(tpaStatus == Action_Up) 
            {
				usbSendSSU(SSU_DATA_UP);
            }
            else if(tpaStatus == Action_Down) 
            {
                usbSendSSU(SSU_DATA_DOWN);
            }
            else if(tpaStatus == Action_Fire) 
            {
				usbSendSSU(SSU_DATA_FIRE);
            }
			
            if(tpaStatus != Action_Idle) 
			{
				monWaitForButtonRelease();
				usbSendSSU(SSU_DATA_STOP);
			}
        }
        else
        {
	        monSetLEDs(Action_NoDev);
            monSync();
        }
    }
}

